@@ -4,26 +4,32 @@ from __future__ import annotations
 import dataclasses as dc
 import hashlib
 import logging
+import os
 import sys
-from typing import Any, cast
-from urllib.parse import urlencode
+from typing import Any
 
 import packaging.requirements
-from packaging.specifiers import SpecifierSet
+from packaging.specifiers import InvalidSpecifier, SpecifierSet
 from packaging.tags import Tag
 from packaging.utils import (
     InvalidWheelFilename,
+    NormalizedName,
     canonicalize_name,
     parse_wheel_filename,
 )
-from packaging.version import InvalidVersion
+from packaging.version import InvalidVersion, Version
 from requests import Session
 
 from unearth.link import Link
 from unearth.pep425tags import get_supported
-from unearth.utils import ARCHIVE_EXTENSIONS, splitext, strip_extras
+from unearth.utils import (
+    ARCHIVE_EXTENSIONS,
+    fix_legacy_specifier,
+    splitext,
+    strip_extras,
+)
 
-logger = logging.getLogger(__package__)
+logger = logging.getLogger(__name__)
 
 
 def is_equality_specifier(specifier: SpecifierSet) -> bool:
@@ -98,23 +104,29 @@ class Package:
 
 @dc.dataclass(frozen=True)
 class FormatControl:
-    only_binary: bool = False
-    no_binary: bool = False
+    only_binary: set[NormalizedName] = dc.field(default_factory=set)
+    no_binary: set[NormalizedName] = dc.field(default_factory=set)
 
-    def __post_init__(self):
-        if self.only_binary and self.no_binary:
-            raise ValueError(
-                "Not allowed to set only_binary and no_binary at the same time."
-            )
+    def get_allowed_formats(self, canonical_name: NormalizedName) -> set[str]:
+        allowed_formats = {"binary", "source"}
+        if canonical_name in self.only_binary:
+            allowed_formats.discard("source")
+        elif canonical_name in self.no_binary:
+            allowed_formats.discard("binary")
+        elif ":all:" in self.only_binary:
+            allowed_formats.discard("source")
+        elif ":all:" in self.no_binary:
+            allowed_formats.discard("binary")
+        return allowed_formats
 
     def check_format(self, link: Link, project_name: str) -> None:
-        if self.only_binary:
-            if not link.is_wheel:
-                raise LinkMismatchError(f"only binaries are allowed for {project_name}")
-        if self.no_binary:
-            if link.is_wheel:
-                raise LinkMismatchError(f"no binary is allowed for {project_name}")
-        return
+        allowed_formats = self.get_allowed_formats(canonicalize_name(project_name))
+        if link.is_wheel and "binary" not in allowed_formats:
+            raise LinkMismatchError(f"binary wheel is not allowed for {project_name}")
+        if not link.is_wheel and "source" not in allowed_formats:
+            raise LinkMismatchError(
+                f"source distribution is not allowed for {project_name}"
+            )
 
 
 @dc.dataclass
@@ -124,16 +136,13 @@ class Evaluator:
     Args:
         package_name (str): The links must match the package name
         target_python (TargetPython): The links must match the target Python
-        hashes (dict[str, list[str]): The links must have the correct hashes
         ignore_compatibility (bool): Whether to ignore the compatibility check
         allow_yanked (bool): Whether to allow yanked candidates
         format_control (bool): Format control flags
     """
 
     package_name: str
-    session: Session
     target_python: TargetPython = dc.field(default_factory=TargetPython)
-    hashes: dict[str, list[str]] = dc.field(default_factory=dict)
     ignore_compatibility: bool = False
     allow_yanked: bool = False
     format_control: FormatControl = dc.field(default_factory=FormatControl)
@@ -141,16 +150,24 @@ class Evaluator:
     def __post_init__(self) -> None:
         self._canonical_name = canonicalize_name(self.package_name)
 
-    def _check_yanked(self, link: Link) -> None:
+    def check_yanked(self, link: Link) -> None:
         if link.yank_reason is not None and not self.allow_yanked:
             yank_reason = f"due to {link.yank_reason}" if link.yank_reason else ""
             raise LinkMismatchError(f"Yanked {yank_reason}")
 
-    def _check_requires_python(self, link: Link) -> None:
+    def check_requires_python(self, link: Link) -> None:
         if not self.ignore_compatibility and link.requires_python:
             py_ver = self.target_python.py_ver or sys.version_info[:2]
             py_version = ".".join(str(v) for v in py_ver)
-            if not SpecifierSet(link.requires_python).contains(py_version, True):
+            try:
+                requires_python = SpecifierSet(
+                    fix_legacy_specifier(link.requires_python)
+                )
+            except InvalidSpecifier:
+                raise LinkMismatchError(
+                    f"Invalid requires-python: {link.requires_python}"
+                )
+            if not requires_python.contains(py_version, True):
                 raise LinkMismatchError(
                     "The target python version({}) doesn't match "
                     "the requires-python specifier {}".format(
@@ -158,53 +175,14 @@ class Evaluator:
                     ),
                 )
 
-    def _check_hashes(self, link: Link) -> None:
-        def hash_mismatch(
-            hash_name: str, given_hash: str, allowed_hashes: list[str]
-        ) -> None:
-            raise LinkMismatchError(
-                f"Hash mismatch, expected: {allowed_hashes}\n"
-                f"got: {hash_name}:{given_hash}"
-            )
-
-        if not self.hashes:
-            return
-        if link.hash_name and link.hash_name in self.hashes:
-            if link.hash not in self.hashes[link.hash_name]:
-                hash_mismatch(
-                    link.hash_name, cast(str, link.hash), self.hashes[link.hash_name]
-                )
-
-        hash_name, allowed_hashes = next(iter(self.hashes.items()))
-        given_hash = self._get_hash(link, hash_name)
-        if given_hash not in allowed_hashes:
-            hash_mismatch(hash_name, given_hash, allowed_hashes)
-
-    def _get_hash(self, link: Link, hash_name: str) -> str:
-        if link.hash_name == hash_name:
-            return cast(str, link.hash)
-        resp = self.session.get(link.normalized, stream=True)
-        hasher = hashlib.new(hash_name)
-        for chunk in resp.iter_content(chunk_size=1024 * 8):
-            hasher.update(chunk)
-        digest = hasher.hexdigest()
-        # Store the hash on the link for future use
-        fragment_dict = link._fragment_dict
-        fragment_dict.pop(link.hash_name, None)  # type: ignore
-        fragment_dict[hash_name] = digest
-        link.__dict__["parsed"] = link.parsed._replace(
-            fragment=urlencode(fragment_dict)
-        )
-        return digest
-
     def evaluate_link(self, link: Link) -> Package | None:
         """
         Evaluate the link and return the package if it matches or None if it doesn't.
         """
         try:
             self.format_control.check_format(link, self.package_name)
-            self._check_yanked(link)
-            self._check_requires_python(link)
+            self.check_yanked(link)
+            self.check_requires_python(link)
             version: str | None = None
             if link.is_wheel:
                 try:
@@ -235,14 +213,39 @@ class Evaluator:
                         raise LinkMismatchError(
                             f"Unsupported archive format: {link.filename}"
                         )
-                version = parse_version_from_egg_info(egg_info, self._canonical_name)
-                if version is None:
+                LOOSE_FILENAME = os.getenv(
+                    "UNEARTH_LOOSE_FILENAME", "false"
+                ).lower() in ("1", "true")
+                if LOOSE_FILENAME:
+                    version = parse_version_from_egg_info(
+                        egg_info, self._canonical_name
+                    )
+                    if version is None:
+                        raise LinkMismatchError(
+                            f"Missing version in the filename: {egg_info}"
+                        )
+                else:
+                    # For source releases, we know that the version should not contain
+                    # any hyphens, so we can split on the last hyphen to get the
+                    # project name.
+                    filename_prefix, has_version, version = egg_info.rpartition("-")
+                    if not has_version:
+                        raise LinkMismatchError(
+                            f"Missing version in the filename: {egg_info}"
+                        )
+                    if canonicalize_name(filename_prefix) != self._canonical_name:
+                        raise LinkMismatchError(
+                            f"The package name doesn't match {egg_info}, set env var "
+                            "UNEARTH_LOOSE_FILENAME=1 to allow legacy filename."
+                        )
+                try:
+                    Version(version)
+                except InvalidVersion:
                     raise LinkMismatchError(
-                        f"Missing version in the filename {egg_info}"
+                        f"Invalid version in the filename {egg_info}: {version}"
                     )
-            self._check_hashes(link)
         except LinkMismatchError as e:
-            logger.debug("Skip link %s: %s", link, e)
+            logger.debug("Skipping link %s: %s", link, e)
             return None
         return Package(name=self.package_name, version=version, link=link)
 
@@ -265,18 +268,49 @@ def evaluate_package(
     if requirement.name:
         if canonicalize_name(package.name) != canonicalize_name(requirement.name):
             logger.debug(
-                "Skip package %s: name doesn't match %s", package, requirement.name
+                "Skipping package %s: name doesn't match %s", package, requirement.name
             )
             return False
 
-    if requirement.specifier and package.version:
-        if not requirement.specifier.contains(
-            package.version, prereleases=allow_prereleases
-        ):
-            logger.debug(
-                "Skip package %s: version doesn't match %s",
-                package,
-                requirement.specifier,
-            )
-            return False
+    if package.version and not requirement.specifier.contains(
+        package.version, prereleases=allow_prereleases
+    ):
+        logger.debug(
+            "Skipping package %s: version doesn't match %s",
+            package,
+            requirement.specifier,
+        )
+        return False
     return True
+
+
+def _get_hash(link: Link, hash_name: str, session: Session) -> str:
+    hasher = hashlib.new(hash_name)
+    with session.get(link.normalized, stream=True) as resp:
+        for chunk in resp.iter_content(chunk_size=1024 * 8):
+            hasher.update(chunk)
+    digest = hasher.hexdigest()
+    if not link.hashes:
+        link.hashes = {}
+    link.hashes[hash_name] = digest
+    return digest
+
+
+def validate_hashes(
+    package: Package, hashes: dict[str, list[str]], session: Session
+) -> bool:
+    if not hashes:
+        return True
+    link = package.link
+    link_hashes = link.hash_option
+    if link_hashes:
+        for hash_name, allowed_hashes in hashes.items():
+            if hash_name in link_hashes:
+                given_hash = link_hashes[hash_name][0]
+                if given_hash not in allowed_hashes:
+                    return False
+                return True
+
+    hash_name, allowed_hashes = next(iter(hashes.items()))
+    given_hash = _get_hash(link, hash_name, session)
+    return given_hash in allowed_hashes
